/*************************************************************************/
/*									 */
/*	Routines to manage tree growth, pruning and evaluation		 */
/*	------------------------------------------------------		 */
/*      MODIFIED BY D. Malerba                                           */
/*									 */
/*************************************************************************/


#include "defns1.i"
#include "types.i"
#include "extern.i"
#include "memory.h"
#include "stdlib.h"

#define MAXCRITICALVALUE 1.0
#define INCREASE         0.01
  
ItemNo		*TargetClassFreq;
Tree		*Raw;
extern Tree	*Pruned;

extern String   ParamFileName;



/****************************************************************************/
/*                                                                          */
/*   DESCRIPTION: Get the value of parameter m from ParamFile to MValues[]  */
/*   AUTHOR: D. Malerba                                                     */
/*                                                                          */
/****************************************************************************/

GetMValues(MValues,NumberOfMValues)
  float   *MValues;
  ItemNo  *NumberOfMValues;

  {
    char string[1000];
    FILE *ParamFile;

    if(!(ParamFile=fopen(ParamFileName,"r")))
      {
        printf("\n\tRead error for file <%s>\n",ParamFileName);
        exit(1);
      };
    while (ReadName(ParamFile,string))
      {
        MValues[*NumberOfMValues] = atof(&string[0]);
        (*NumberOfMValues)++;
        if (*NumberOfMValues > 1000)
          {
            printf("\n\tToo much values for m in the file <%s>\n",ParamFileName);
            fclose(ParamFile);
            exit(1);
          }
        if (*NumberOfMValues <= 0)
          {
            printf("\n\tNo values for m in the file <%s>\n",ParamFileName);
            fclose(ParamFile);
            exit(1);
          }
      };
    if (*NumberOfMValues <= 0)
      {
        printf("\n\tNo values for m in the file <%s>\n",ParamFileName);
        fclose(ParamFile);
        exit(1);
      };
    fclose(ParamFile);
  } 



/*************************************************************************/
/*                                                                       */
/*    DESCRIPTION:                                                       */
/*      Computes the leaf number and the error rate for a                */
/*      pruned tree Pruned[h] and saves these statististics              */
/*      in StatPrunedTree[h].                                            */
/*    AUTHOR: D. Malerba                                                 */
/*                                                                       */
/*************************************************************************/

EvaluatePrunedTree(T,Leaf,Error)
  Tree  T;
  int   *Leaf;
  float *Error;

  {
    ClassNo RealClass,PrunedClass,Category();
    ItemNo i,PrunedErrors=0;

    ForEach(i, 0, MaxItem)
      {
        RealClass = Class(Item[i]);
	PrunedClass = Category(Item[i],T);
	if ( PrunedClass != RealClass ) PrunedErrors++;
      };  
    *Leaf = LeafCount(T);
    *Error = 100.0 * PrunedErrors/(MaxItem+1.0);
 } 


/*************************************************************************/
/*                                                                       */
/*   DESCRIPTION:                                                        */
/*           Write in T->Upper and T->Lower the value                    */ 
/*           of alfa for the node T and return the minimum alfa          */
/*   AUTHOR: D. Malerba                                                  */
/*                                                                       */
/*************************************************************************/

UpdateAlfaTree(T,R,MinAlfa)
  Tree   T,R;
  float  *MinAlfa;

  {
    int i;
    float   Alfa;

    if (T->NodeType)
      {
        Alfa= ((T->Errors - R->Errors)/(MaxItem+1))/(LeafCount(T)-1);
        if (Alfa<=*MinAlfa) *MinAlfa=Alfa;
        T->Upper=Alfa;  /* Upper and Lower must be equal, otherwise are considered soften threshold */ 
        T->Lower=Alfa;
        ForEach(i,1,T->Forks)
          UpdateAlfaTree(T->Branch[i],R->Branch[i],MinAlfa);
      }
  }


/*************************************************************************/
/*                                                                       */
/*    DESCRIPTION: find the M-th tree with minimum size and error rate   */
/*    AUTHOR: D. Malerba                                                 */
/*                                                                       */
/*************************************************************************/

Minimum(trees,a,No,M)
  Tree  *trees;
  float *a;
  int   No,*M;
  
  {
    int i,LeafCount();
    float minsize,minerror;

    minsize=LeafCount(trees[0]);
    minerror=a[0];
    *M=0;
    ForEach(i,1,No) 
      if (a[i] < minerror)
      {
       	minerror = a[i];
    	minsize = LeafCount(trees[i]);
        *M = i;
      }
    else if (minerror == a[i])
           if (LeafCount(trees[i]) <= minsize)
             { 
     		minerror = a[i];
                minsize = LeafCount(trees[i]);
                *M = i;
             }
  }



/****************************************************************************/
/*      								    */
/*	DESCRIPTION:                                                        */
/*          Grow and prune a single tree with one of ten method of pruning  */
/*      AUTHOR: D. Malerba                                                  */
/*                                                                          */
/****************************************************************************/


OneTree(PruningMethod)
  short  PruningMethod;

/*  ---------  */

{
    ItemNo   UnprunedErrors,NumberOfMValues=0,Misclassify(),MisclassifyForIGP();
    Tree     FormTree(), CopyTree(),*TreeToPrune, ExpandTree();
    Boolean  Prune();
    int      NodeNumber=1,MaxNodeNumber=1,i,Best=0,Step=1;
    float    *MValues,BestCriticalValue=0.0;
    int      LeafTreePruned,MinSize;                 /* Leaf number and error rate for pruned trees */  
    float    ErrorTreePruned,MinError;               /* Error rate for pruned trees */
    Tree     *MinPruned,*PrecTree,*Grown;     
    float    CriticalV=INCREASE;                     /* Max value for the gain ratio measure */ 
    Boolean  *PruneDone=false,CriticalValuePruning(),ReducedErrorPruning(),MinimumErrorPruning(),PessimisticErrorPruning1SE(), GraftingOrPruning();
    int      TreeNumber=1,LeafCount(),NumberOfAlfaValues=0;
    int      v,M;
    float    *AlfaValues,MinAlfaValue=1000.0,Threshold;
    float    *ErrorP,StandardDeviation;

    InitialiseTreeData();
    InitialiseWeights();

    Raw = (Tree *) calloc(1, sizeof(Tree));
    Pruned = (Tree *) calloc(1000, sizeof(Tree));
    MinPruned = (Tree *) calloc(1,sizeof(Tree));
    PrecTree  = (Tree *) calloc(1,sizeof(Tree));
    Grown  = (Tree *) calloc(1,sizeof(Tree));

    AllKnown = true;
    Raw[0] = FormTree(0, MaxItem);
    printf("\n");
    
    switch (PruningMethod)
      {
        case GRA               :
        case SE0               :
        case SE1               :
        case REP               :                                         /* pruning methods using growing set */
        case CVP               :
        case MEP               : printf("GROWN TREE\n\n");
                                 PrintTree(Raw[0]);
                                 SaveTree(Raw[0],".tree_grow");
                                 Pruned[0] = CopyTree(Raw[0]);
                                 switch (PruningMethod)
                                   {
                                     case SE0:
                                     case SE1: Pruned[1]=CopyTree(Raw[0]);
                                               UpdateTree(Pruned[1],0,MaxItem);
                                               AlfaValues=(float *) calloc(1000,sizeof(float));
                                               TreeToPrune=(Tree *) calloc(1,sizeof(Tree));
                                               AlfaValues[0]=0.0;
                                               TreeToPrune[0]=CopyTree(Raw[0]);
                                                Verbosity(1)    
                                                  {
                                                    printf("\n\n\nTREE NUMBER 1 \tAlfa=0.0\n\n");
                                                    PrintTree(Pruned[0]);
                                                  }
                                               v=2;
                                               while (LeafCount(Pruned[v-1])!=1)
                                                 {
                                                   UpdateAlfaTree(Pruned[v-1],TreeToPrune[0],&MinAlfaValue);
                                                   AlfaValues[v-1]=MinAlfaValue;
                                                   Verbosity(1)    
                                                     printf("\n\n\nTREE NUMBER %d \tAlfa=%.6f\n\n",v,MinAlfaValue);
                                                   PruneAlfa(MinAlfaValue,Pruned[v-1],TreeToPrune[0],NodeNumber,&MaxNodeNumber);
                                                   Verbosity(1)
                                                       PrintTree(Pruned[v-1]);
                                                   Pruned[v]=CopyTree(Pruned[v-1]);
                                                   v++;
                                                   if (v>1000)
                                                     {
                                                       printf("\n\n\nError -> family of pruned trees too large\n\n");
                                                       exit(1);
                                                     }
                                                   MaxNodeNumber=1;
                                                   MinAlfaValue=1000.0; 
                                                 }
                                               ReleaseTree(TreeToPrune[0]);
                                               NumberOfAlfaValues=v-1;
                                               ErrorP=(float *) calloc (NumberOfAlfaValues,sizeof(float));
                                               FreeItem();
                                               GetData(".prune");
                                               ForEach(v,0,NumberOfAlfaValues-1)
                                                   EvaluatePrunedTree(Pruned[v],&LeafTreePruned,&ErrorP[v]);
                                               printf("\n\n");
                                               ForEach(v,0,NumberOfAlfaValues-1)
                                                 printf("TREE NUMBER %d Alfa=%.6f \tEstimated Error rate: %.2f %s\n",v+1,AlfaValues[v],ErrorP[v],"%");
                                               Minimum(Pruned,ErrorP,NumberOfAlfaValues-1,&M);
                                               if (PruningMethod==SE1)
                                                 {
                                                   StandardDeviation=sqrt(ErrorP[M]*(100-ErrorP[M])/(MaxItem+1));
                                                   Threshold=ErrorP[M] + StandardDeviation;
                                                   printf("\n\nBEST TREE NUMBER %d ERROR: %.2f \tSTANDARD DEVIATION= %.4f\n\n",M+1,ErrorP[M],StandardDeviation);
                                                   ForEach(i,0,NumberOfAlfaValues-1) 
                                                     if (ErrorP[i] <= Threshold) M=i;
                                                   printf("\n\n\nPRUNED TREE (ERROR COMPLEXITY PRUNING 1SE PRUNING SET) Number %d\n\n",M+1);
                                                 }   
                                               else printf("\n\n\nPRUNED TREE (ERROR COMPLEXITY PRUNING 0SE PRUNING SET) Number %d\n\n",M+1);
                                               if (M>0)
                                                 {
                                                   ReleaseTree(Pruned[0]);
                                                   Pruned[0]=CopyTree(Pruned[M]);
                                                 }
                                               PrintTree(Pruned[0]);
                                               ForEach(i,1,NumberOfAlfaValues-1)
                                                 ReleaseTree(Pruned[i]);
                                               cfree(ErrorP);
                                               cfree(AlfaValues);
                                               FreeItem();
                                               GetData(".grow");
                                               break;
                                     case REP: GetData(".prune");                      /* read the pruning set */
                                               UnprunedErrors = Misclassify(Raw[0]);   /* number of errors for the grown tree */
                                               Verbosity(1)
                                                 {  
                                                   TreeToPrune=(Tree *) calloc(1, sizeof(Tree));
                                                   TreeToPrune[0]=CopyTree(Raw[0]);

                                                   ReallocWeight();   
                                                   UpdateTree(TreeToPrune[0],0,MaxItem);  /* update a tree with new data */ 
             
                                                   printf("\nTREE TO PRUNE:\n\n");
                                                   PrintTree(TreeToPrune[0]);
             
                                                   ReleaseTree(TreeToPrune[0]);
                                                 }
                                                ReducedErrorPruning(Pruned[0],&UnprunedErrors,NodeNumber,&MaxNodeNumber);
                                                printf("\n\n\nPRUNED TREE (REDUCED ERROR PRUNING)\n\n");
                                                PrintTree(Pruned[0]);
                                                Verbosity(1)
                                                  {
                                                    printf("\n\nEvaluation on pruning data (%d items): \n",MaxItem+1);
                                                    Evaluate(false,0);
                                                  };
                                                GetData(".grow");
                                                break;
                                     case GRA:  FreeItem();  
                                                GetData(".prune");
                                                UnprunedErrors=Misclassify(Raw[0]);
                                                
GraftingOrPruning(Pruned[0],&UnprunedErrors,NodeNumber,&MaxNodeNumber);
                                                printf("\n\n\nPRUNED OR GRAFTED TREE\n\n");
                                                PrintTree(Pruned[0]);
                                                Verbosity(1)
                                                   {
                                                     printf("\n\nEvaluation on pruning data (%d items): \n",MaxItem+1);
                                                     Evaluate(false,0);
                                                    };
                                                FreeItem(); 
                                                GetData(".grow"); 
                                                break;
                                     case CVP:  UpdateTree(Pruned[0],0,MaxItem);  /* update a tree with new data */
                                                GetData(".prune");
                                                EvaluatePrunedTree(Pruned[0],&LeafTreePruned,&ErrorTreePruned);
                                                MinError = ErrorTreePruned;
                                                MinSize = LeafTreePruned;
                                                MinPruned[0] = CopyTree(Pruned[0]);
                                                Verbosity(1)
                                                  {
                                                     printf("\n\n\nTREE NUMBER %d \tCV=%.2f (CRITICAL VALUE PRUNING METHOD)\n",TreeNumber++,BestCriticalValue);
                                                     printf("\nLeaf number: %d \tError rate: %.2f\n\n",LeafTreePruned,ErrorTreePruned);  
                                                     PrintTree(Pruned[0]);
                                                  } 

                                                      /* Check for the best pruned tree */                                                                 
                                                while (!((CriticalV > (MAXCRITICALVALUE+Epsilon)) || (LeafCount(Pruned[0])==1)))
                                                  {
                                                    CriticalValuePruning(CriticalV,Pruned[0],NodeNumber,&MaxNodeNumber,&PruneDone);
                                                    if (PruneDone)
                                                      {
                                                         Verbosity(1)
                                                           printf("\n\n\nTREE NUMBER %d \tCV=%.2f (CRITICAL VALUE PRUNING METHOD)\n",TreeNumber++,CriticalV);
                                                         EvaluatePrunedTree(Pruned[0],&LeafTreePruned,&ErrorTreePruned);
                                                         Verbosity(1)
                                                           {
                                                             printf("\nLeaf number: %d \tError rate: %.2f\n\n",LeafTreePruned,ErrorTreePruned);  
                                                             PrintTree(Pruned[0]);
                                                           } 
                                                         if (ErrorTreePruned < MinError)
                                                           {
                                                	     MinError = ErrorTreePruned;
                                               		     MinSize = LeafTreePruned;
                                                             ReleaseTree(MinPruned[0]);
                                                             MinPruned[0] = CopyTree(Pruned[0]);
                                                             BestCriticalValue=CriticalV;
                                                           }
                                                         else if (MinError == ErrorTreePruned)
                                                                if (LeafTreePruned <= MinSize)
                                                                  { 
                                                		     MinError = ErrorTreePruned;
                                               			     MinSize = LeafTreePruned;
                                                        	     ReleaseTree(MinPruned[0]);
                                                                     MinPruned[0] = CopyTree(Pruned[0]);
                                                               	     BestCriticalValue=CriticalV;
                                                                  }
                                                       }
                                                     MaxNodeNumber = 1;
                                                     PruneDone = false;
                                                     CriticalV+=INCREASE;
                                                  }
                                                printf("\n\n\nBEST PRUNED TREE (CRITICAL VALUE PRUNING METHOD)\n");
                                                printf("FOR cv=%.2f\n\n",BestCriticalValue);
                                                ReleaseTree(Pruned[0]);
                                                Pruned[0] = CopyTree(MinPruned[0]);
                                                ReleaseTree(MinPruned[0]);
                                                PrintTree(Pruned[0]);
                                                GetData(".grow"); 
                                                break;
                                     case MEP : GetData(".prune");
                                                MValues = (float *) calloc(1000,sizeof(float));
                                                GetMValues(MValues,&NumberOfMValues);
                                                Verbosity(1) printf("\nNUMBER OF VALUES OF m IS: %d\n\n",NumberOfMValues);
                                                
                                                Verbosity(1)
                                                  printf("\n\n\nTREE NUMBER 1 \tM=%f (MINIMUM ERROR PRUNING METHOD)\n",MValues[0]);
                                                MinimumErrorPruning(MValues[0],Pruned[0],NodeNumber,&MaxNodeNumber);
                                                EvaluatePrunedTree(Pruned[0],&LeafTreePruned,&ErrorTreePruned);
                                                MinPruned[0]=CopyTree(Pruned[0]);
                                                Verbosity(1)
                                                  {
                                                    printf("\nLeaf number: %d \tError rate: %.2f\n\n",LeafTreePruned,ErrorTreePruned);  
                                                    PrintTree(Pruned[0]);
                                                  }; 
                                                MaxNodeNumber=1;
                                                MinError = ErrorTreePruned;
                                                MinSize = LeafTreePruned;
                                                
                                                ForEach(i,1,NumberOfMValues-1)
                                                  {
                                                    ReleaseTree(Pruned[0]);
                                                    Pruned[0]=CopyTree(Raw[0]);
                                                    Verbosity(1)
                                                      printf("\n\n\nTREE NUMBER %d \tM=%f (MINIMUM ERROR PRUNING METHOD)\n",i+1,MValues[i]);
                                                    MinimumErrorPruning(MValues[i],Pruned[0],NodeNumber,&MaxNodeNumber);
                                                    EvaluatePrunedTree(Pruned[0],&LeafTreePruned,&ErrorTreePruned);
                                                    Verbosity(1)
                                                      {
                                                         printf("\nLeaf number: %d \tError rate: %.2f\n\n",LeafTreePruned,ErrorTreePruned);  
                                                         PrintTree(Pruned[0]);
                                                      };
                                                    if (ErrorTreePruned < MinError)
                                                      {
                                                	MinError = ErrorTreePruned;
                                               		MinSize = LeafTreePruned;
                                                        ReleaseTree(MinPruned[0]);
                                                        MinPruned[0] = CopyTree(Pruned[0]);
                                                        Best=i;
                                                      }
                                                    else if (MinError == ErrorTreePruned)
                                                           if (LeafTreePruned <= MinSize)
                                                             { 
                                                		MinError = ErrorTreePruned;
                                               			MinSize = LeafTreePruned;
                                                        	ReleaseTree(MinPruned[0]);
                                                        	MinPruned[0] = CopyTree(Pruned[0]);
                                                               	Best=i;
                                                             };
                                                    MaxNodeNumber = 1;
                                                  };
                                                printf("\n\n\nBEST PRUNED TREE (MINIMUM ERROR PRUNING METHOD)\n");
                                                printf("FOR m=%f\n\n",MValues[Best]);
                                                ReleaseTree(Pruned[0]);
                                                Pruned[0] = CopyTree(MinPruned[0]);
                                                ReleaseTree(MinPruned[0]);
                                                PrintTree(Pruned[0]);
                                                GetData(".grow"); 
                                                break;
                                   } 
                                 break;
         case IGP              : if (Raw[0]->NodeType)
                                   ForEach(i,1,Raw[0]->Forks)
                                     ClassLabels(Raw[0]->Branch[i],Raw[0]->Leaf);
                                 printf("\n\n\n\t\t\tSTEP NUMBER %d (ITERATIVE GROWING AND PRUNING METHOD)\n",Step++);
                                 printf("\n\nGROWN TREE\n\n"); 
                                 PrintTree(Raw[0]);
                                 SaveTree(Raw[0],".tree_grow");
                                 Pruned[0] = CopyTree(Raw[0]);
                                 Grown[0]  = CopyTree(Raw[0]);
                                 FreeItem();
                                 GetData(".train2");                      /* read the pruning set */
                                 printf("\nRed %d cases from file <%s.train2> for pruning\n\n",MaxItem+1,FileName);
                                 UnprunedErrors = MisclassifyForIGP(Grown[0]);    /* number of errors for the grown tree */
                                 ReducedErrorPruningForIGP(Pruned[0],&UnprunedErrors,NodeNumber,&MaxNodeNumber);
                                 printf("\n\n\nPRUNED TREE\n\n");
                                 PrintTree(Pruned[0]);
                                 
                                 do
                                   {
                                     PrecTree[0]=CopyTree(Pruned[0]);
                                     ReleaseTree(Grown[0]);
                                     Grown[0]=CopyTree(Pruned[0]);
                                     ReleaseTree(Pruned[0]);
                                     MaxNodeNumber=1;
 
                                          /* Expand tree */
                             
       
                                     ReallocWeight();
                                     Grown[0] = ExpandTree(Grown[0],0,MaxItem);
                                                                               	                                                                                                        printf("\n\n\n\t\t\tSTEP NUMBER %d (ITERATIVE GROWING AND PRUNING METHOD)\n",Step++);
                                     printf("\n\nGROWN TREE\n\n");
                                     PrintTree(Grown[0]);
                                     Pruned[0]=CopyTree(Grown[0]);
 
                                        /* Prune Tree */
                                     
                                     FreeItem();
                                     if ((Step-1)%2 == 0) 
                                       {
                                         GetData(".train1");
                                         printf("\nRed %d cases from file <%s.train1> for pruning\n\n",MaxItem+1,FileName);
                                       }
                                     else 
                                       {
                                         GetData(".train2");
                                         printf("\nRed %d cases from file <%s.train2> for pruning\n\n",MaxItem+1,FileName);
                                       }
                                     UnprunedErrors = MisclassifyForIGP(Grown[0]);    /* number of errors for the expanded tree */
                                     ReducedErrorPruningForIGP(Pruned[0],&UnprunedErrors,NodeNumber,&MaxNodeNumber);
                                     printf("\n\n\nPRUNED TREE\n\n");
                                     PrintTree(Pruned[0]);
                                   }
                                 while (LeafCount(PrecTree[0])!=LeafCount(Pruned[0])); 
                                 
                                 printf("\n\n\nBEST TREE (ITERATIVE GROWING AND PRUNING METHOD)\n");
                                 printf("AT STEP = %d\n\n",Step-1);
                                 PrintTree(Pruned[0]);
                                 GetData(".train"); 
                                 break;
         case SE2              : 
         case SE3              : 
         case SE4              : 
         case SE5              : 
         case PEP              : printf("\n\nTRAINED TREE\n\n"); 
                                 PrintTree(Raw[0]);
                                 SaveTree(Raw[0],".tree_train");
                                 Pruned[0] = CopyTree(Raw[0]);
                                 switch(PruningMethod)
                                   {
                                     case PEP : TreeToPrune=(Tree *) calloc(1, sizeof(Tree));
                                                TreeToPrune[0]=CopyTree(Raw[0]);
                                                UpdateTree(TreeToPrune[0],0,MaxItem);         /* update a tree with new data */ 
                                                PessimisticErrorPruning1SE(Pruned[0],TreeToPrune[0],NodeNumber,&MaxNodeNumber);
                                                printf("\n\n\nPRUNED TREE (PESSIMISTIC ERROR PRUNING 1SE)\n\n");
                                                PrintTree(Pruned[0]);
                                                ReleaseTree(TreeToPrune[0]); 
                                                break;
                                     case SE4 : 
                                     case SE5 : 
                                     case SE2 :
                                     case SE3 : Pruned[1]=CopyTree(Raw[0]);
                                                UpdateTree(Pruned[1],0,MaxItem);
                                                AlfaValues=(float *) calloc(1000,sizeof(float));
                                                TreeToPrune=(Tree *) calloc(1,sizeof(Tree));
                                                TreeToPrune[0]=CopyTree(Raw[0]);
                                                AlfaValues[0]=0.0;   
                                                Verbosity(1)    
                                                  {
                                                    printf("\n\n\nTREE NUMBER 1 \tAlfa=0.0\n\n");
                                                    PrintTree(Pruned[0]);
                                                  }
                                                v=2;
                                                while (LeafCount(Pruned[v-1])!=1)
                                                  {
                                                    UpdateAlfaTree(Pruned[v-1],TreeToPrune[0],&MinAlfaValue);
                                                    if (MinAlfaValue >0)
                                                    {
                                                     AlfaValues[v-1]=MinAlfaValue;
                                                     Verbosity(1);    
                                                      printf("\n\n\nTREE NUMBER %d \tAlfa=%.6f\n\n",v,MinAlfaValue);
                                                     PruneAlfa(MinAlfaValue,Pruned[v-1],TreeToPrune[0],NodeNumber,&MaxNodeNumber);
                                                     Verbosity(1)
                                                        PrintTree(Pruned[v-1]);
                                                     Pruned[v]=CopyTree(Pruned[v-1]);
                                                     v++;
                                                     if (v>1000)
                                                      {
                                                        printf("\n\n\nError -> family of pruned trees too large\n\n");
                                                        exit(1);
                                                      }
                                                    }
                                                    else
                                                      PruneAlfa(MinAlfaValue,Pruned[v-1],TreeToPrune[0],NodeNumber,&MaxNodeNumber);
                                                    MaxNodeNumber=1;
                                                    MinAlfaValue=1000.0; 
                                                  }
                                                ReleaseTree(TreeToPrune[0]);
                                                NumberOfAlfaValues=v-1;
                                                ErrorP=(float *) calloc (NumberOfAlfaValues,sizeof(float));
                                               switch(PruningMethod)
                                               {
                                                case SE2:
                                                case SE3:CrossValidation(ErrorP,AlfaValues,NumberOfAlfaValues-1);
                                                case SE4:
                                                case SE5:CrossValidationNew(ErrorP,AlfaValues,NumberOfAlfaValues-1);
                                               }
                                                printf("\n\n");
                                                printf("\nAlfa Values for TRAINED TREE Tmax\n\n");
                                                ForEach(v,0,NumberOfAlfaValues-1)
                                                  printf("TREE NUMBER %d Alfa=%.6f \tEstimated Error rate: %.2f\n",v+1,AlfaValues[v],ErrorP[v]);
                                                Minimum(Pruned,ErrorP,NumberOfAlfaValues-1,&M);
                                                if ((PruningMethod==SE3)|| (PruningMethod==SE5))
                                                  {
                                                    StandardDeviation=sqrt(ErrorP[M]*(1-ErrorP[M])/(Raw[0]->Items));
                                                    Threshold=ErrorP[M] + StandardDeviation;
                                                    printf("\n\nBEST TREE NUMBER %d ERROR: %.2f \tSTANDARD DEVIATION= %.4f\n\n",M+1,ErrorP[M],StandardDeviation);
                                                    ForEach(i,0,NumberOfAlfaValues-1) 
                                                      if (ErrorP[i] <= Threshold) M=i;
                                                    printf("\n\n\nPRUNED TREE (ERROR COMPLEXITY PRUNING 1SE CV) Number %d\n\n",M+1);
                                                  }   
                                                else printf("\n\n\nPRUNED TREE (ERROR COMPLEXITY PRUNING 0SE CV) Number %d\n\n",M+1);
                                                if (M>0)
                                                  {
                                                    ReleaseTree(Pruned[0]);
                                                    Pruned[0]=CopyTree(Pruned[M]);
                                                  }
                                                PrintTree(Pruned[0]);
                                                ForEach(i,1,NumberOfAlfaValues-1)
                                                  ReleaseTree(Pruned[i]);
                                                break;
                                   }
                                 break;
         case StandardPrune    : PrintTree(Raw[0]); 
                                 SaveTree(Raw[0], ".unpruned");
 
                                 Pruned[0] = CopyTree(Raw[0]);
                                 if ( Prune(Pruned[0]) )
                                   {
	                             printf("\nSimplified ");
	                             PrintTree(Pruned[0]);
                                   }
                                 break;
  }
}



/*************************************************************************/
/*									 */
/*	Grow and prune TRIALS trees and select the best of them		 */
/*									 */
/*************************************************************************/


short BestTree()
/*    --------  */
{
    Tree CopyTree(), Iterate();
    Boolean Prune();
    short t, Best=0;

    InitialiseTreeData();

    TargetClassFreq = (ItemNo *) calloc(MaxClass+1, sizeof(ItemNo));

    Raw    = (Tree *) calloc(TRIALS, sizeof(Tree));
    Pruned = (Tree *) calloc(TRIALS, sizeof(Tree));

    /*  If necessary, set initial size of window to 20% (or twice
	the sqrt, if this is larger) of the number of data items,
	and the maximum number of items that can be added to the
	window at each iteration to 20% of the initial window size  */

    if ( ! WINDOW )
    {
	WINDOW = Max(2 * sqrt(MaxItem+1.0), (MaxItem+1) / 5);
    }

    if ( ! INCREMENT )
    {
	INCREMENT = Max(WINDOW / 5, 1);
    }

    FormTarget(WINDOW);

    /*  Form set of trees by iteration and prune  */

    ForEach(t, 0, TRIALS-1 )
    {
        FormInitialWindow();

	printf("\n--------\nTrial %d\n--------\n\n", t);

	Raw[t] = Iterate(WINDOW, INCREMENT);
	printf("\n");
	PrintTree(Raw[t]);

	SaveTree(Raw[t], ".unpruned");

	Pruned[t] = CopyTree(Raw[t]);
	if ( Prune(Pruned[t]) )
	{
	    printf("\nSimplified ");
	    PrintTree(Pruned[t]);
	}

	if ( Pruned[t]->Errors < Pruned[Best]->Errors )
	{
	    Best = t;
	}
    }
    printf("\n--------\n");

    return Best;
}



/*************************************************************************/
/*									 */
/*  The windowing approach seems to work best when the class		 */
/*  distribution of the initial window is as close to uniform as	 */
/*  possible.  FormTarget generates this initial target distribution,	 */
/*  setting up a TargetClassFreq value for each class.			 */
/*									 */
/*************************************************************************/


    FormTarget(Size)
/*  -----------  */
    ItemNo Size;
{
    ItemNo i, *ClassFreq;
    ClassNo c, Smallest, ClassesLeft=0;

    ClassFreq = (ItemNo *) calloc(MaxClass+1, sizeof(ItemNo));

    /*  Generate the class frequency distribution  */

    ForEach(i, 0, MaxItem)
    {
	ClassFreq[ Class(Item[i]) ]++;
    }

    /*  Calculate the no. of classes of which there are items  */

    ForEach(c, 0, MaxClass)
    {
	if ( ClassFreq[c] )
	{
	    ClassesLeft++;
	}
	else
	{
	    TargetClassFreq[c] = 0;
	}
    }

    while ( ClassesLeft )
    {
	/*  Find least common class of which there are some items  */

	Smallest = -1;
	ForEach(c, 0, MaxClass)
	{
	    if ( ClassFreq[c] &&
		 ( Smallest < 0 || ClassFreq[c] < ClassFreq[Smallest] ) )
	    {
		Smallest = c;
	    }
	}

	/*  Allocate the no. of items of this class to use in the window  */

	TargetClassFreq[Smallest] = Min(ClassFreq[Smallest], Round(Size/ClassesLeft));

	ClassFreq[Smallest] = 0;

	Size -= TargetClassFreq[Smallest];
	ClassesLeft--;
    }

    cfree(ClassFreq);
}



/*************************************************************************/
/*									 */
/*  Form initial window, attempting to obtain the target class profile	 */
/*  in TargetClassFreq.  This is done by placing the targeted number     */
/*  of items of each class at the beginning of the set of data items.	 */
/*									 */
/*************************************************************************/


    FormInitialWindow()
/*  -------------------  */
{
    ItemNo i, Start=0, More;
    ClassNo c;
    void Swap();

    Shuffle();

    ForEach(c, 0, MaxClass)
    {
	More = TargetClassFreq[c];

	for ( i = Start ; More ; i++ )
	{
	    if ( Class(Item[i]) == c )
	    {
		Swap(Start, i);
		Start++;
		More--;
	    }
	}
    }
}



/*************************************************************************/
/*									 */
/*		Shuffle the data items randomly				 */
/*									 */
/*************************************************************************/


    Shuffle()
/*  -------  */
{
    ItemNo This, Alt, Left;
    Description Hold;

    This = 0;
    for( Left = MaxItem+1 ; Left ; )
    {
        Alt = This + (Left--) * Random;
        Hold = Item[This];
        Item[This++] = Item[Alt];
        Item[Alt] = Hold;
    }
}



/*************************************************************************/
/*									 */
/*  Grow a tree iteratively with initial window size Window and		 */
/*  initial window increment IncExceptions.				 */
/*									 */
/*  Construct a classifier tree using the data items in the		 */
/*  window, then test for the successful classification of other	 */
/*  data items by this tree.  If there are misclassified items,		 */
/*  put them immediately after the items in the window, increase	 */
/*  the size of the window and build another classifier tree, and	 */
/*  so on until we have a tree which successfully classifies all	 */
/*  of the test items or no improvement is apparent.			 */
/*									 */
/*  On completion, return the tree which produced the least errors.	 */
/*									 */
/*************************************************************************/


Tree Iterate(Window, IncExceptions)
/*   -------  */
    ItemNo Window, IncExceptions;
{
    Tree Classifier, BestClassifier=Nil, FormTree();
    ItemNo i, Errors, TotalErrors, BestTotalErrors=MaxItem+1,
	   Exceptions, Additions;
    ClassNo Assigned, Category();
    short Cycle=0;
    void Swap();

    printf("Cycle   Tree    -----Cases----");
    printf("    -----------------Errors-----------------\n");
    printf("        size    window   other");
    printf("    window  rate   other  rate   total  rate\n");
    printf("-----   ----    ------  ------");
    printf("    ------  ----  ------  ----  ------  ----\n");

    do
    {
	/*  Build a classifier tree with the first Window items  */

	InitialiseWeights();
	AllKnown = true;
	Classifier = FormTree(0, Window-1);

	/*  Error analysis  */

	Errors = Round(Classifier->Errors);

	/*  Move all items that are incorrectly classified by the
	    classifier tree to immediately after the items in the
	    current window.  */

	Exceptions = Window;
	ForEach(i, Window, MaxItem)
	{
	    Assigned = Category(Item[i], Classifier);
	    if ( Assigned != Class(Item[i]) )
	    {
		Swap(Exceptions, i);
		Exceptions++;
	    }
	}
        Exceptions -= Window;
	TotalErrors = Errors + Exceptions;

	/*  Print error analysis  */

	printf("%3d  %7d  %8d  %6d  %8d%5.1f%%  %6d%5.1f%%  %6d%5.1f%%\n",
	       ++Cycle, TreeSize(Classifier), Window, MaxItem-Window+1,
	       Errors, 100*(float)Errors/Window,
	       Exceptions, 100*Exceptions/(MaxItem-Window+1.001),
	       TotalErrors, 100*TotalErrors/(MaxItem+1.0));

	/*  Keep track of the most successful classifier tree so far  */

	if ( ! BestClassifier || TotalErrors < BestTotalErrors )
	{
	    if ( BestClassifier ) ReleaseTree(BestClassifier);
	    BestClassifier = Classifier;
	    BestTotalErrors = TotalErrors;
        }
	else
	{
	    ReleaseTree(Classifier);
	}

	/*  Increment window size  */

	Additions = Min(Exceptions, IncExceptions);
	Window = Min(Window + Max(Additions, Exceptions / 2), MaxItem + 1);
    }
    while ( Exceptions );

    return BestClassifier;
}




/*****************************************************************************/
/*                                                                           */
/*      DESCRPTION: Return the number of misclassifications in the tree T    */
/*                                                                           */
/*****************************************************************************/


ItemNo Misclassify(T)
    Tree T;
{    
   ItemNo Errors,i;
   ClassNo RealClass,Category(); 
 
   Errors = 0;
   ForEach(i, 0, MaxItem)
     {
       RealClass = Class(Item[i]);
       if ( Category(Item[i], T) != RealClass ) Errors++;
     }
   return Errors;
}                             



/*************************************************************************/
/*                                                                       */
/*      DESCRIPTION:                                                     */
/*           Save on files tree001 prune001 the tree size and the        */ 
/*           percentage of errors for unpruned tree and pruned tree,     */
/*           respectivement.                                             */
/*      AUTHOR: D. Malerba                                               */
/*                                                                       */
/*************************************************************************/


EvaluateAndSave(Saved)
  short Saved;

{
    ClassNo RealClass,Row,Col, PrunedClass, Category();
    short t;
    ItemNo *ConfusionMat,i, RawErrors, PrunedErrors,*ElementsForClass,Count;
    FILE   *PruneStatFile,*TreeStatFile,*MatDistFile;

    ConfusionMat = (ItemNo *) calloc((MaxClass+1)*(MaxClass+1), sizeof(ItemNo));
    ElementsForClass = (ItemNo *) calloc(MaxClass+1,sizeof(ItemNo));

    ForEach(t, 0, TRIALS-1)
    {
	RawErrors = PrunedErrors = 0;

	ForEach(i, 0, MaxItem)
	{
	    RealClass = Class(Item[i]);

	    if ( Category(Item[i], Raw[t]) != RealClass ) RawErrors++;

	    PrunedClass = Category(Item[i], Pruned[t]);

	    if ( PrunedClass != RealClass ) PrunedErrors++;

            ConfusionMat[RealClass*(MaxClass+1)+PrunedClass]++;

	}
    
        /*   Saves statistics on the tree001 and prune001 files    */

        if ((TreeStatFile = fopen("tree001","a")) != NULL) 
          fprintf(TreeStatFile,"\t%9d\t%6.2f\t\n",LeafCount(Raw[t]),100.0*RawErrors / (MaxItem + 1.0));
        else
          {
            printf("\nCan't open the TreeStatFile tree001\n\n");
            exit(1);
          };
        if ((PruneStatFile = fopen("prune001","a")) != NULL)
          fprintf(PruneStatFile,"\t%9d\t%6.2f\t\n",LeafCount(Pruned[t]),100.0*PrunedErrors / (MaxItem+1.0));
        else
          {
            printf("\nCan't open the PruneStatFile Prune001\n\n");
            exit(1);
          };  
     }  
     
     /* Save distribuction matrix on MatDist001 file */

     if ((MatDistFile = fopen("MatDist001","a")) != NULL)
       {
         fprintf(MatDistFile,"%c\n",'*');
         
         ForEach(Row,0,MaxClass)
           {
             Count=0;
             ForEach(Col,0,MaxClass)
               Count+=ConfusionMat[Row*(MaxClass+1)+Col];
             ElementsForClass[Row] = Count;
           }

          ForEach(Row,0,MaxClass)
           {  
             ForEach(Col,0,MaxClass)
               fprintf(MatDistFile,"\t%6.2f",(ElementsForClass[Row] ? (ConfusionMat[Row*(MaxClass+1)+Col]*100.0)/ElementsForClass[Row] : 0.00));
             fprintf(MatDistFile,"\n");
           };
       }     
     else
       {
         printf("\nCan't open MatDistFile MatDist001\n\n");
         exit(1);
       };
     free(ConfusionMat); 
}

  
/*************************************************************************/
/*									 */
/*	Print report of errors for each of the trials			 */
/*									 */
/*************************************************************************/


    Evaluate(CMInfo, Saved)
/*  --------  */
    Boolean CMInfo;
    short Saved;
{
    ClassNo RealClass, PrunedClass, Category();
    short t;
    ItemNo *ConfusionMat, i, RawErrors, PrunedErrors;
    FILE   *PruneStat,*TreeStat;

    if ( CMInfo )
    {
	ConfusionMat = (ItemNo *) calloc((MaxClass+1)*(MaxClass+1), sizeof(ItemNo));
    }

    printf("\n");

    if ( TRIALS > 1 )
    {
	printf("Trial\t Before Pruning           After Pruning\n");
	printf("-----\t----------------   ---------------------------\n");
    }
    else
    {
	printf("\t Before Pruning           After Pruning\n");
	printf("\t----------------   ---------------------------\n");
    }
    printf("\tSize      Errors   Size      Errors   Estimate\n\n");

    ForEach(t, 0, TRIALS-1)
    {
	RawErrors = PrunedErrors = 0;

	ForEach(i, 0, MaxItem)
	{
	    RealClass = Class(Item[i]);

	    if ( Category(Item[i], Raw[t]) != RealClass ) RawErrors++;

	    PrunedClass = Category(Item[i], Pruned[t]);

	    if ( PrunedClass != RealClass ) PrunedErrors++;

	    if ( CMInfo && t == Saved )
	    {
		ConfusionMat[RealClass*(MaxClass+1)+PrunedClass]++;
	    }
	}
    
	if ( TRIALS > 1 )
	{
	    printf("%4d", t);
	}

	printf("\t%4d  %3d(%4.1f%%)   %4d  %3d(%4.1f%%)    (%4.1f%%)%s\n",
	       TreeSize(Raw[t]), RawErrors, 100.0*RawErrors / (MaxItem+1.0),
	       TreeSize(Pruned[t]), PrunedErrors, 100.0*PrunedErrors / (MaxItem+1.0),
	       100 * Pruned[t]->Errors / Pruned[t]->Items,
	       ( t == Saved ? "   <<" : "" ));
    }

    if ( CMInfo )
    {
	PrintConfusionMatrix(ConfusionMat);
	free(ConfusionMat);
    }
}
