/*************************************************************************/
/*                                                              	 */
/*  DESCRIPTION:                                                         */
/*       Determine the class of a case description from a decision tree	 */
/*       The class is not the first better like in classify.c, but the   */
/*       one of the label of the leaf.                                   */
/*       Attention, this is rationable only for classification of        */
/*       Known cases.                                                    */
/*  AUTHOR: D. Malerba                                                   */
/*                                                              	 */
/*************************************************************************/


#include "defns1.i"
#include "types.i"
#include "extern.i"

extern Tree *Pruned,*Raw;
float	    *ClassSum1=Nil;		/* ClassSum1[c] = total weight of class c */


/*************************************************************************/
/*                                                              	 */
/*  DESCRIPTION:                                                         */
/*       Categorize a case description using the given decision tree	 */
/*  AUTHOR: D. Malerba                                                   */
/*                                                              	 */
/*************************************************************************/


ClassNo CategoryForIGP(CaseDesc, DecisionTree) 
/*       --------  */
    Description CaseDesc; 
    Tree DecisionTree; 
{ 
    ClassNo c, BestClass;

    if ( ! ClassSum1 )
    {
	ClassSum1 = (float *) malloc((MaxClass+1) * sizeof(float));
    }

    ForEach(c, 0, MaxClass)
    {
	ClassSum1[c] = 0;
    }

    ClassifyForIGP(CaseDesc, DecisionTree, 1.0);

    BestClass = 0;
    ForEach(c, 0, MaxClass)
    {
	Verbosity(5) printf("class %s weight %.2f\n", ClassName[c], ClassSum1[c]);

	if ( ClassSum1[c] > ClassSum1[BestClass] ) BestClass = c;
    }
    
    return BestClass;
} 



/****************************************************************************/
/*                                                                          */
/*  DESCRIPTION:                                                            */
/*       Classify a case description using the given subtree by adjusting   */
/*       the value ClassSum for each class.			            */
/*       For unknown cases it compute the probability classes distribution  */
/*       while for known cases (Weight=1.0) return probability 1 for the    */
/*       class that labels the leaf reached and probability 0 for the others*/ 
/*  AUTHOR: D. Malerba                                                      */
/*                                                              	    */
/****************************************************************************/


   ClassifyForIGP(CaseDesc, T, Weight)
/*  --------  */
    Description CaseDesc; 
    Tree T;
    float Weight;
{
    DiscrValue v, dv;
    float Cv;
    Attribute a;
    ClassNo c;

    switch ( T->NodeType )
    {
        case 0:  /* leaf */

	    if (( T->Items > 0 )&&(Weight<1.0))
	    {
		/*  Update from ALL classes  */

		ForEach(c, 0, MaxClass)
		{
		    if ( T->ClassDist[c] )
		    {
			ClassSum1[c] += Weight * T->ClassDist[c] / T->Items;
		    }
		}
	    }
	    else
	    {
		ClassSum1[T->Leaf] += Weight;
	    }

	    return;

	case BrDiscr:  /* test of discrete attribute */

	    a = T->Tested;
	    v = DVal(CaseDesc, a);

	    if ( v && v <= T->Forks )	/*  Make sure not new discrete value  */
	    {
		ClassifyForIGP(CaseDesc, T->Branch[v], Weight);
	    }
	    else
	    {
		ForEach(v, 1, T->Forks)
		{
		   ClassifyForIGP(CaseDesc, T->Branch[v], 
			     (Weight * T->Branch[v]->Items) / T->Items);
		}
	    }

	    return;

	case ThreshContin:  /* test of continuous attribute */

	    a = T->Tested;
	    Cv = CVal(CaseDesc, a);

	    if ( Cv == Unknown )
	    {
		ForEach(v, 1, 2)
		{
		    ClassifyForIGP(CaseDesc, T->Branch[v], 
			     (Weight * T->Branch[v]->Items) / T->Items);
		}
	    }
	    else
	    {
		v = ( Cv <= T->Cut ? 1 : 2 );
		ClassifyForIGP(CaseDesc, T->Branch[v], Weight);
	    }

	    return;

	case BrSubset:  /* subset test on discrete attribute  */

	    a = T->Tested;
	    dv = DVal(CaseDesc, a);

	    if ( dv )
	    {
		ForEach(v, 1, T->Forks)
		{
		    if ( In(dv, T->Subset[v]) )
		    {
			ClassifyForIGP(CaseDesc, T->Branch[v], Weight);

			return;
		    }
		}
	    }

	    /*  Value unknown or not found in any of the subsets  */

	    ForEach(v, 1, T->Forks)
	    {
		ClassifyForIGP(CaseDesc, T->Branch[v], 
		         (Weight * T->Branch[v]->Items) / T->Items);
	    }

	    return;
    } 
}



/*************************************************************************/
/*                                                                       */
/*    DESCRIPTION:                                                       */
/*         Prune with the reduced error pruning method,                  */ 
/*         return true if the tree is changed                            */
/*    AUTHOR: D. Malerba                                                 */
/*                                                                       */
/*************************************************************************/

Boolean ReducedErrorPruningForIGP(T,UnprunedErrors,NodeNumber,MaxNodeNumber)
   Tree   T;
   ItemNo *UnprunedErrors;
   int    NodeNumber;
   int    *MaxNodeNumber;
  {
    Boolean Changed;
    short   PreviousNodeType;
    ItemNo  PrunedErrors,i,v,MisclassifyForIGP();
    int     PreviousNodeNumber;

    Changed=false;
    if (T->NodeType != 0)
      {
        PreviousNodeNumber=NodeNumber;
        ForEach(i,1,T->Forks)
          {
            (*MaxNodeNumber)++;
            NodeNumber= (*MaxNodeNumber);
            ReducedErrorPruningForIGP(T->Branch[i],UnprunedErrors,NodeNumber,MaxNodeNumber);
          };
        PreviousNodeType = T->NodeType;                        /*   Save information for the current node T   */
 
        NodeNumber=PreviousNodeNumber;
  
        /*  Modify the current node T in a leaf  */

        T->NodeType = 0;
        PrunedErrors = MisclassifyForIGP(Pruned[0]);
        if (*UnprunedErrors >= PrunedErrors)
          {
            ForEach(v,1,T->Forks)
              {
                ReleaseTree(T->Branch[v]);
                T->Branch[v] = Nil;
              };
            Verbosity(1)
                printf("\nPRUNED NODE NUMBER: %d \tPRUNED ERRORS: %d \tNOT PRUNED ERRORS: %d",NodeNumber,PrunedErrors,*UnprunedErrors);
            *UnprunedErrors = PrunedErrors;
            T->Errors=(T->Items) - (T->ClassDist[T->Leaf]);
            Changed = true;
          }
        else 
          { 
            printf("\nnot PRUNED NODE NUMBER: %d \tPRUNED ERRORS: %d \tNOT PRUNED ERRORS: %d",NodeNumber,PrunedErrors,*UnprunedErrors);
            T->NodeType = PreviousNodeType;
          } 
      }
    else (*MaxNodeNumber)--;
    return Changed;
  } 


/************************************************************************/
/*                                                                      */
/*      DESCRIPTION:                                                    */
/*           Return the number of misclassifications in the tree T      */
/*      AUTHOR: D. Malerba                                              */
/*                                                                      */
/************************************************************************/


ItemNo MisclassifyForIGP(T)
    Tree T;
{    
   ItemNo Errors,i;
   ClassNo RealClass,CategoryForIGP(); 
 
   Errors = 0;
   ForEach(i, 0, MaxItem)
     {
       RealClass = Class(Item[i]);
       if ( CategoryForIGP(Item[i], T) != RealClass ) Errors++;
     }
   return Errors;
}                             


/*************************************************************************/
/*                                                                       */
/*      DESCRIPTION:                                                     */
/*           Save on files tree001 prune001 the tree size and the        */ 
/*           percentage of errors for unpruned tree and pruned tree,     */
/*           respectivement.                                             */
/*      AUTHOR: D. Malerba                                               */
/*                                                                       */
/*************************************************************************/


EvaluateAndSaveForIGP(Saved)
  short Saved;

{
    ClassNo RealClass,Row,Col, PrunedClass, CategoryForIGP();
    short t;
    ItemNo *ConfusionMat,i, RawErrors, PrunedErrors,*ElementsForClass,Count;
    FILE   *PruneStatFile,*TreeStatFile,*MatDistFile;

    ConfusionMat = (ItemNo *) calloc((MaxClass+1)*(MaxClass+1), sizeof(ItemNo));
    ElementsForClass = (ItemNo *) calloc(MaxClass+1,sizeof(ItemNo));

    ForEach(t, 0, TRIALS-1)
    {
	RawErrors = PrunedErrors = 0;

	ForEach(i, 0, MaxItem)
	{
	    RealClass = Class(Item[i]);

	    if ( CategoryForIGP(Item[i], Raw[t]) != RealClass ) RawErrors++;

	    PrunedClass = CategoryForIGP(Item[i], Pruned[t]);

	    if ( PrunedClass != RealClass ) PrunedErrors++;

            ConfusionMat[RealClass*(MaxClass+1)+PrunedClass]++;

	}
    
        /*   Saves statistics on the tree001 and prune001 files    */

        if ((TreeStatFile = fopen("tree001","a")) != NULL) 
          fprintf(TreeStatFile,"\t%9d\t%6.2f\t\n",LeafCount(Raw[t]),100.0*RawErrors / (MaxItem + 1.0));
        else
          {
            printf("\nCan't open the TreeStatFile tree001\n\n");
            exit(1);
          };
        if ((PruneStatFile = fopen("prune001","a")) != NULL)
          fprintf(PruneStatFile,"\t%9d\t%6.2f\t\n",LeafCount(Pruned[t]),100.0*PrunedErrors / (MaxItem+1.0));
        else
          {
            printf("\nCan't open the PruneStatFile Prune001\n\n");
            exit(1);
          };  
     }  
     
     /* Save distribuction matrix on MatDist001 file */

     if ((MatDistFile = fopen("MatDist001","a")) != NULL)
       {
         fprintf(MatDistFile,"%c\n",'*');
         
         ForEach(Row,0,MaxClass)
           {
             Count=0;
             ForEach(Col,0,MaxClass)
               Count+=ConfusionMat[Row*(MaxClass+1)+Col];
             ElementsForClass[Row] = Count;
           }

          ForEach(Row,0,MaxClass)
           {  
             ForEach(Col,0,MaxClass)
               fprintf(MatDistFile,"\t%6.2f",(ElementsForClass[Row] ? (ConfusionMat[Row*(MaxClass+1)+Col]*100.0)/ElementsForClass[Row] : 0.00));
             fprintf(MatDistFile,"\n");
           };
       }     
     else
       {
         printf("\nCan't open MatDistFile MatDist001\n\n");
         exit(1);
       };
     free(ConfusionMat); 
}

  
/*************************************************************************/
/*									 */
/*	DESCRIPTION:                                                     */
/*           Print report of errors for each of the trials	         */
/*      AUTHOR: D. Malerba                                               */
/*									 */
/*************************************************************************/


    EvaluateForIGP(CMInfo, Saved)
/*  --------  */
    Boolean CMInfo;
    short Saved;
{
    ClassNo RealClass, PrunedClass, CategoryForIGP();
    short t;
    ItemNo *ConfusionMat, i, RawErrors, PrunedErrors;
    FILE   *PruneStat,*TreeStat;

    if ( CMInfo )
    {
	ConfusionMat = (ItemNo *) calloc((MaxClass+1)*(MaxClass+1), sizeof(ItemNo));
    }

    printf("\n");

    if ( TRIALS > 1 )
    {
	printf("Trial\t Before Pruning           After Pruning\n");
	printf("-----\t----------------   ---------------------------\n");
    }
    else
    {
	printf("\t Before Pruning           After Pruning\n");
	printf("\t----------------   ---------------------------\n");
    }
    printf("\tSize      Errors   Size      Errors   Estimate\n\n");

    ForEach(t, 0, TRIALS-1)
    {
	RawErrors = PrunedErrors = 0;

	ForEach(i, 0, MaxItem)
	{
	    RealClass = Class(Item[i]);

	    if ( CategoryForIGP(Item[i], Raw[t]) != RealClass ) RawErrors++;

	    PrunedClass = CategoryForIGP(Item[i], Pruned[t]);

	    if ( PrunedClass != RealClass ) PrunedErrors++;

	    if ( CMInfo && t == Saved )
	    {
		ConfusionMat[RealClass*(MaxClass+1)+PrunedClass]++;
	    }
	}
    
	if ( TRIALS > 1 )
	{
	    printf("%4d", t);
	}

	printf("\t%4d  %3d(%4.1f%%)   %4d  %3d(%4.1f%%)    (%4.1f%%)%s\n",
	       TreeSize(Raw[t]), RawErrors, 100.0*RawErrors / (MaxItem+1.0),
	       TreeSize(Pruned[t]), PrunedErrors, 100.0*PrunedErrors / (MaxItem+1.0),
	       100 * Pruned[t]->Errors / Pruned[t]->Items,
	       ( t == Saved ? "   <<" : "" ));
    }

    if ( CMInfo )
    {
	PrintConfusionMatrix(ConfusionMat);
	free(ConfusionMat);
    }
}
