 /*************************************************************************/
/*									 */
/*	Prune a decision tree and predict its error rate		 */
/*	------------------------------------------------		 */
/*									 */
/*************************************************************************/


#include "defns1.i"
#include "types.i"
#include "extern.i"
#include "stdlib.h"


extern	ItemCount	*Weight;
extern  Tree            *Pruned,*Raw;

Set	*PossibleValues=Nil;
Boolean	Changed;

#define	LocalVerbosity(x)	if (Sh >= 0 && VERBOSITY >= x)
#define	Intab(x)		Indent(x, "| ")



/*************************************************************************/
/*									 */
/*  Prune tree T, returning true if tree has been modified		 */
/*									 */
/*************************************************************************/


Boolean Prune(T)
/*      -----  */
    Tree T;
{
    ItemNo i;
    float EstimateErrors();
    Attribute a;

    InitialiseWeights();
    AllKnown = true;

    Verbosity(1) printf("\n");

    Changed = false;

    EstimateErrors(T, 0, MaxItem, 0, true);

    if ( SUBSET )
    {
	if ( ! PossibleValues )
	{
	    PossibleValues = (Set *) calloc(MaxAtt+1, sizeof(Set));
	}

	ForEach(a, 0, MaxAtt)
	{
	    if ( MaxAttVal[a] )
	    {
		PossibleValues[a] = (Set) malloc((MaxAttVal[a]>>3) + 1);
		ClearBits((MaxAttVal[a]>>3) + 1, PossibleValues[a]);
		ForEach(i, 1, MaxAttVal[a])
		{
		    SetBit(i, PossibleValues[a]);
		}
	    }
	}

	CheckPossibleValues(T);
    }

    return Changed;
}



/*************************************************************************/
/*                                                                       */
/*       DESCRIPTION: prune T if its alfa value is Alfa                  */
/*       AUTHOR: D. Malerba                                              */
/*                                                                       */
/*************************************************************************/

PruneAlfa(Alfa,T,R,NodeNumber,MaxNodeNumber)
  float Alfa;
  Tree  T,R;
  int   NodeNumber,*MaxNodeNumber;

  {
    Boolean Changed;
    int     PreviousNodeNumber,v,i;
    float   SubTreeErrors=0.0;

    Changed=false;
    if (T->NodeType != 0)
      {
        PreviousNodeNumber=NodeNumber;
        ForEach(i,1,T->Forks)
          {
            (*MaxNodeNumber)++;
            NodeNumber= (*MaxNodeNumber);
            PruneAlfa(Alfa,T->Branch[i],R->Branch[i],NodeNumber,MaxNodeNumber);
            SubTreeErrors+=R->Branch[i]->Errors;
          };
        NodeNumber=PreviousNodeNumber;
 
        if (T->Upper==Alfa)
          {
            Verbosity(1)
                printf("PRUNED NODE NUMBER: %d \tR(t): %.2f \tR(Tt): %.2f \tAlfa: %.7f\n",NodeNumber,(T->Errors)/(MaxItem+1),(R->Errors)/(MaxItem+1),Alfa);
            ForEach(v,1,T->Forks)
              {
                ReleaseTree(T->Branch[v]);
                T->Branch[v] = Nil;
                ReleaseTree(R->Branch[v]);
                R->Branch[v]=Nil;
              };
            R->Errors=T->Errors;
            T->NodeType=0;
            R->NodeType=0;
            Changed = true;
          }
        else R->Errors=SubTreeErrors;
      }
    else (*MaxNodeNumber)--;
    return Changed;
  } 


 
/*************************************************************************/
/*                                                                       */
/*      DESCRIPTION: Compute the expected error rate for the node T      */
/*      AUTHOR: D. Malerba                                               */
/*                                                                       */
/*************************************************************************/

float EER(MValue,T)
  float MValue;
  Tree  T;

  {
    float  BestEER=0,TempEER;
    ItemNo c;

    TempEER = (T->Items - T->ClassDist[0] + (1-(Raw[0]->ClassDist[0]/Raw[0]->Items))*MValue)/(T->Items+MValue);
    BestEER = TempEER;
    ForEach(c,1,MaxClass)
      {
        TempEER = (T->Items - T->ClassDist[c] + (1-(Raw[0]->ClassDist[c]/Raw[0]->Items))*MValue)/(T->Items+MValue);
        if (TempEER<=BestEER) BestEER=TempEER; 
      }
    return BestEER; 
  }


/************************************************************************/
/*                                                                      */
/*       DESCRPTION: Prune with the critical value pruning method,      */
/*                   return true if the tree is changed                 */
/*       AUTHOR: D. Malerba                                             */
/*                                                                      */
/************************************************************************/

Boolean CriticalValuePruning(CriticalV,T,NodeNumber,MaxNodeNumber,PruneDone)
  float    CriticalV;
  Tree     T;
  int      NodeNumber;
  int      *MaxNodeNumber;
  Boolean  *PruneDone;

  {
    Boolean PrunedChild=true;
    ItemNo  i,v;
    int     PreviousNodeNumber;
    float   GainRatio;

    if (T->NodeType != 0)
      {
        GainRatio = T->Upper;
        PreviousNodeNumber = NodeNumber;
        ForEach(i,1,T->Forks)
          {
            (*MaxNodeNumber)++;
            NodeNumber = (*MaxNodeNumber);
            PrunedChild=(CriticalValuePruning(CriticalV,T->Branch[i],NodeNumber,MaxNodeNumber,PruneDone) && PrunedChild);
          }
        NodeNumber = PreviousNodeNumber;
  
        if ((GainRatio <= (CriticalV+Epsilon)) && (PrunedChild))
          {
            T->NodeType = 0;
            ForEach(v,1,T->Forks)
              { 
                ReleaseTree(T->Branch[v]);
                T->Branch[v] = Nil;
              };
            *PruneDone = true;
            Verbosity(1)
                printf("\nPRUNED NODE NUMBER: %d \tGAIN-RATIO: %.3f \tTHRESHOLD: %.2f",NodeNumber,GainRatio,CriticalV);
          }
        else
          { 
            PrunedChild=false;
          } 
      }
    else (*MaxNodeNumber)--;
    return PrunedChild;
  } 


/*************************************************************************/
/*									 */
/*       DESCRIPTION: Prune with the minimum error pruning method,       */
/*                    return true if the tree is changed                 */
/*       AUTHOR: D. Malerba                                              */
/*                                                                       */
/*************************************************************************/

Boolean MinimumErrorPruning(MValue,T,NodeNumber,MaxNodeNumber)
  float  MValue;
  Tree   T;
  int    NodeNumber;
  int    *MaxNodeNumber;

  {
    Boolean Changed;
    ItemNo  i,v;
    int     PreviousNodeNumber;
    float   ExpectedErrorRate,DynamicError=0.0,EER();

    Changed=false;
    T->Errors = EER(MValue,T);                     /* compute the static error for the node T */
    if (T->NodeType != 0)
      {
        PreviousNodeNumber=NodeNumber;
        ForEach(i,1,T->Forks)
          {
            (*MaxNodeNumber)++;
            NodeNumber= (*MaxNodeNumber);
            MinimumErrorPruning(MValue,T->Branch[i],NodeNumber,MaxNodeNumber);
          };
        NodeNumber=PreviousNodeNumber;
  
        ForEach(i,1,T->Forks)                      /* compute the dinamic error for the node T */
          DynamicError+=(T->Branch[i]->Errors * (T->Branch[i]->Items)/(T->Items));
        if (T->Errors <= DynamicError)
          {
            T->NodeType=0;
            ForEach(v,1,T->Forks)
              { 
                ReleaseTree(T->Branch[v]);
                T->Branch[v] = Nil;
              };
            Verbosity(1)
                printf("\nPRUNED NODE NUMBER: %d \tSE(t): %.2f \tDE(t): %.2f",NodeNumber,T->Errors*100.0,DynamicError*100.0);
            Changed = true;
          }
        else 
          {  
             printf("\nNot Pruned Node Number: %d \tSE(t): %.2f \tDE(t): %.2f",NodeNumber,T->Errors*100.0,DynamicError*100.0);  
             T->Errors = DynamicError;                        /* adjust the static error for the node T */ 
          }
      }
    else (*MaxNodeNumber)--;
    return Changed;
  } 



/*************************************************************************/
/*                                                                       */
/*       DESCRIPTION:Prune with the reduced error pruning method,        */ 
/*                   return true if the tree is changed                  */
/*                                                                       */
/*************************************************************************/

Boolean ReducedErrorPruning(T,UnprunedErrors,NodeNumber,MaxNodeNumber)
   Tree   T;
   ItemNo *UnprunedErrors;
   int    NodeNumber;
   int    *MaxNodeNumber;
  {
    Boolean Changed;
    short   PreviousNodeType;
    ItemNo  PrunedErrors,i,v,Misclassify();
    int     PreviousNodeNumber;

    Changed=false;
    if (T->NodeType != 0)
      {
        PreviousNodeNumber=NodeNumber;
        ForEach(i,1,T->Forks)
          {
            (*MaxNodeNumber)++;
            NodeNumber= (*MaxNodeNumber);
            ReducedErrorPruning(T->Branch[i],UnprunedErrors,NodeNumber,MaxNodeNumber);
          };
        PreviousNodeType = T->NodeType;                        /*   Save information for the current node T   */
 
        NodeNumber=PreviousNodeNumber;
  
        /*  Modify the current node T in a leaf  */

        T->NodeType = 0;
        PrunedErrors = Misclassify(Pruned[0]);
        if (*UnprunedErrors >= PrunedErrors)
          {
            ForEach(v,1,T->Forks)
              {
                ReleaseTree(T->Branch[v]);
                T->Branch[v] = Nil;
              };
            Verbosity(1)
              printf("\nPRUNED NODE NUMBER: %d \tPRUNED ERRORS: %d \tNOT PRUNED ERRORS: %d",NodeNumber,PrunedErrors,*UnprunedErrors);
            *UnprunedErrors = PrunedErrors;
            T->Errors=(T->Items) - (T->ClassDist[T->Leaf]);
            Changed = true;
          }
        else 
          { 
            printf("\nnot PRUNED NODE NUMBER: %d \tPRUNED ERRORS: %d \tNOT PRUNED ERRORS: %d",NodeNumber,PrunedErrors,*UnprunedErrors);
            T->NodeType = PreviousNodeType;
          } 
      }
    else (*MaxNodeNumber)--;
    return Changed;
  } 


/*****************************************************************************/
/*                                                                           */
/*       DESCRIPTION: Prune with the pessimistic pruning method, 1SE rule,   */ 
/*           return true if the tree is changed                              */
/*       AUTHOR: D. Malerba                                                  */
/*                                                                           */
/*****************************************************************************/

Boolean PessimisticErrorPruning1SE(T,UpdateT,NodeNumber,MaxNodeNumber)
   Tree   T,UpdateT;
   int    NodeNumber;
   int    *MaxNodeNumber;
  {
    Boolean Changed;
    int     i,v;
    float   NodeErrors,SubTreeErrors,StandardError;

    Changed=false;
    if (T->NodeType != 0)
      {
        NodeErrors= (UpdateT->Errors + 0.5);                                      /* compute n'(t) */
        SubTreeErrors= (T->Errors + ((float) LeafCount(T))/2);                    /* compute n'(Tt) */
        StandardError=sqrt(SubTreeErrors*(T->Items - SubTreeErrors)/T->Items);
        
        if (NodeErrors <= (SubTreeErrors + StandardError))
          {
            AdjustMaxNodeNumber(T,MaxNodeNumber);
            T->NodeType=0;
            T->Errors=UpdateT->Errors;
            ForEach(v,1,T->Forks)
              {
                ReleaseTree(T->Branch[v]);
                T->Branch[v] = Nil;
              };
            Verbosity(1)
                printf("\nPRUNED NODE NUMBER: %d \tN(t): %.2f \tn'(t): %.2f \tn'(Tt): %.2f \tSE: %.2f",NodeNumber,T->Items,NodeErrors,SubTreeErrors,StandardError);
            Changed = true;
          }
        else
          ForEach(i,1,T->Forks)
            {
              (*MaxNodeNumber)++;
              NodeNumber = (*MaxNodeNumber);
              PessimisticErrorPruning1SE(T->Branch[i],UpdateT->Branch[i],NodeNumber,MaxNodeNumber);
            };
      }
    else (*MaxNodeNumber)--;
    return Changed;
  } 


/*************************************************************************/
/*                                                                       */
/*         DESCRIPTION: scan subtree for adjusting MaxNodeNumber         */
/*         AUTHOR: D. Malerba                                            */
/*                                                                       */
/*************************************************************************/

int AdjustMaxNodeNumber(T,MaxNodeNumber)
    Tree T;
    int  *MaxNodeNumber;
  {
    int i;
    
    if (T->NodeType!=0)
      {
         ForEach(i,1,T->Forks)
           {
             (*MaxNodeNumber)++;
             AdjustMaxNodeNumber(T->Branch[i],MaxNodeNumber);
           }
      }
    else (*MaxNodeNumber)--;
  }
    
/*************************************************************************/
/*									 */
/*	Estimate the errors in a given subtree				 */
/*									 */
/*************************************************************************/

 
float EstimateErrors(T, Fp, Lp, Sh, UpdateTree)
/*    --------------  */
    Tree T;
    ItemNo Fp, Lp; 
    short Sh;
    Boolean UpdateTree;
{ 
    ItemNo i, Kp, Ep, Group();
    ItemCount Cases, KnownCases, *LocalClassDist, TreeErrors, LeafErrors,
	ExtraLeafErrors, BranchErrors, CountItems(), Factor, MaxFactor, AddErrs();
    DiscrValue v, MaxBr;
    ClassNo c, BestClass;
    Boolean PrevAllKnown;

    /*  Generate the class frequency distribution  */

    Cases = CountItems(Fp, Lp);
    LocalClassDist = (ItemCount *) calloc(MaxClass+1, sizeof(ItemCount));
    
    ForEach(i, Fp, Lp)
    { 
	LocalClassDist[ Class(Item[i]) ] += Weight[i];
    } 

    /*  Find the most frequent class and update the tree  */

    BestClass = T->Leaf;
    ForEach(c, 0, MaxClass)
    {
	if ( LocalClassDist[c] > LocalClassDist[BestClass] )
	{
	    BestClass = c;
	}
    }
    LeafErrors = Cases - LocalClassDist[BestClass];
    ExtraLeafErrors = AddErrs(Cases, LeafErrors);

    if ( UpdateTree )
    {
	T->Items = Cases;
	T->Leaf  = BestClass;
	memcpy(T->ClassDist, LocalClassDist, (MaxClass + 1) * sizeof(ItemCount));
    }

    if ( ! T->NodeType )	/*  leaf  */
    {
	TreeErrors = LeafErrors + ExtraLeafErrors;

	if ( UpdateTree )
	{
	    T->Errors = TreeErrors;

	    LocalVerbosity(1)
	    {
		Intab(Sh);
	    	printf("%s (%.2f:%.2f/%.2f)\n", ClassName[T->Leaf],
	    		T->Items, LeafErrors, T->Errors);
	    }
	}

	free(LocalClassDist);

	return TreeErrors;
    }

    /*  Estimate errors for each branch  */

    Kp = Group(0, Fp, Lp, T) + 1;
    KnownCases = CountItems(Kp, Lp);

    PrevAllKnown = AllKnown;
    if ( Kp != Fp ) AllKnown = false;

    TreeErrors = MaxFactor = 0;

    ForEach(v, 1, T->Forks)
    {
	Ep = Group(v, Kp, Lp, T);

	if ( Kp <= Ep )
	{
	    Factor = CountItems(Kp, Ep) / KnownCases;

	    if ( Factor >= MaxFactor )
	    {
		MaxBr = v;
		MaxFactor = Factor;
	    }

	    ForEach(i, Fp, Kp-1)
	    {
		Weight[i] *= Factor;
	    }

	    TreeErrors += EstimateErrors(T->Branch[v], Fp, Ep, Sh+1, UpdateTree);

	    Group(0, Fp, Ep, T);
	    ForEach(i, Fp, Kp-1)
	    {
		Weight[i] /= Factor;
	    }
	}
    }
 
    AllKnown = PrevAllKnown;

    if ( ! UpdateTree )
    {
	free(LocalClassDist);

	return TreeErrors;
    }

    /*  See how the largest branch would fare  */

    BranchErrors = EstimateErrors(T->Branch[MaxBr], Fp, Lp, -1000, false);

    LocalVerbosity(1)
    {
        Intab(Sh);
        printf("%s:  [%d%%  N=%.2f  tree=%.2f  leaf=%.2f+%.2f  br[%d]=%.2f]\n",
		AttName[T->Tested],
		(int) ((TreeErrors * 100) / (T->Items + 0.001)),
		T->Items, TreeErrors, LeafErrors, ExtraLeafErrors,
		MaxBr, BranchErrors);
    }

    /*  See whether tree should be replaced with leaf or largest branch  */

    if ( LeafErrors + ExtraLeafErrors <= BranchErrors + 0.1 &&
	 LeafErrors + ExtraLeafErrors <= TreeErrors + 0.1 )
    {
	LocalVerbosity(1)
	{
	    Intab(Sh);
	    printf("Replaced with leaf %s\n", ClassName[T->Leaf]);
	}

	T->NodeType = 0;
	T->Errors = LeafErrors + ExtraLeafErrors;
	Changed = true;
    }
    else
    if ( BranchErrors <= TreeErrors + 0.1 )
    {
	LocalVerbosity(1)
	{
	    Intab(Sh);
	    printf("Replaced with branch %d\n", MaxBr);
	}

	AllKnown = PrevAllKnown;
	EstimateErrors(T->Branch[MaxBr], Fp, Lp, Sh, true);
	memcpy((char *) T, (char *) T->Branch[MaxBr], sizeof(TreeRec));
	Changed = true;
    }
    else
    {
	T->Errors = TreeErrors;
    }

    AllKnown = PrevAllKnown;
    free(LocalClassDist);

    return T->Errors;
}



/*************************************************************************/
/*									 */
/*	Remove unnecessary subset tests on missing values		 */
/*									 */
/*************************************************************************/


    CheckPossibleValues(T)
/*  -------------------  */
    Tree T;
{
    Set HoldValues;
    int v, Bytes, b;
    Attribute A;
    char Any=0;

    if ( T->NodeType == BrSubset )
    {
	A = T->Tested;

	Bytes = (MaxAttVal[A]>>3) + 1;
	HoldValues = (Set) malloc(Bytes);

	/*  See if last (default) branch can be simplified or omitted  */

	ForEach(b, 0, Bytes-1)
	{
	    T->Subset[T->Forks][b] &= PossibleValues[A][b];
	    Any |= T->Subset[T->Forks][b];
	}

	if ( ! Any )
	{
	    T->Forks--;
	}

	/*  Process each subtree, leaving only values in branch subset  */

	CopyBits(Bytes, PossibleValues[A], HoldValues);

	ForEach(v, 1, T->Forks)
	{
	    CopyBits(Bytes, T->Subset[v], PossibleValues[A]);

	    CheckPossibleValues(T->Branch[v]);
	}

	CopyBits(Bytes, HoldValues, PossibleValues[A]);

	free(HoldValues);
    }
    else
    if ( T->NodeType )
    {
	ForEach(v, 1, T->Forks)
	{
	    CheckPossibleValues(T->Branch[v]);
	}
    }
}







Boolean GraftingOrPruning(T,UnprunedErrors,NodeNumber,MaxNodeNumber)
Tree T;
ItemNo *UnprunedErrors;
int NodeNumber;
int *MaxNodeNumber;
{
 Tree Temp;
 Boolean Changed,Graft;
 short PreviousNodeType,GraftedNodeType;
 ItemNo PrunedErrors,GraftedErrors=123945687,k,GraftedErrorsTemp,i,v,Misclassify(),j;
 ClassNo c,BestClass;
 int PreviousNodeNumber,*MemNode;
 
 

Changed=false;
if(T->NodeType!=0)
 {
   /* Temp=(Tree) calloc(1,sizeof(Tree)); */
   if((MemNode=(int*) calloc(T->Forks, sizeof(int)))==NULL) exit(1);
   PreviousNodeNumber=NodeNumber;
   ForEach(i,1,T->Forks)
     {
       (*MaxNodeNumber)++;
       NodeNumber=(*MaxNodeNumber);
       MemNode[i-1]=NodeNumber;
       GraftingOrPruning(T->Branch[i],UnprunedErrors,NodeNumber,MaxNodeNumber);
     }
       Temp=(Tree) calloc(1,sizeof(Tree));
       Graft=false;
       PreviousNodeType=T->NodeType;
       NodeNumber=PreviousNodeNumber;
       k=0;
       ForEach(j,1,T->Forks)
       {
        if((T->Branch[j]!=Nil)&&(T->Branch[j]->NodeType!=0))
        {
         Graft=true;
         memcpy((char *)Temp,(char *) T,sizeof(TreeRec));
         memcpy((char *)T,(char *) T->Branch[j],sizeof(TreeRec));
         GraftedErrorsTemp=Misclassify(Pruned[0]);
         if(GraftedErrorsTemp<=GraftedErrors)
            {
              GraftedErrors=GraftedErrorsTemp;
              k=j;
             }
         memcpy((char *)T,(char *) Temp,sizeof(TreeRec));
         }
        }
        GraftedNodeType=T->NodeType; 
        T->NodeType=0;
        PrunedErrors=Misclassify(Pruned[0]);
        if((PrunedErrors<=*UnprunedErrors) &&((Graft==false)||(PrunedErrors<=GraftedErrors))) 
          {
               ForEach(v,1,T->Forks)
               {
                 ReleaseTree(T->Branch[v]);
                 T->Branch[v]=Nil;
               }
               T->NodeType=0;
               T->Errors=(T->Items)-(T->ClassDist[T->Leaf]);
               Verbosity(1);
               printf("\nNode number: %d \tErrors when pruned: %d \tErrors when not pruned: %d \tPRUNED",NodeNumber,PrunedErrors,*UnprunedErrors);
               *UnprunedErrors=PrunedErrors;
               Changed=true;
             }
             else
             {
               if((GraftedErrors<=*UnprunedErrors)&&(Graft==true)&&(GraftedErrors<=PrunedErrors))
                  { 
                     memcpy((char *)Temp,(char *) T,sizeof(TreeRec)); 
                     memcpy((char *)T,(char *) T->Branch[k],sizeof(TreeRec));
                     T->NodeType=GraftedNodeType;
                     T->Items=Temp->Items;
                     T->ClassDist=Temp->ClassDist;
                     T->Errors=(T->Items)-(T->ClassDist[T->Leaf]);
                     /* FreeItem();
                     GetData(".grow");
                     AdjustTreeGrafted(T);
                     FreeItem();
                     GetData(".prune"); */  
                     Verbosity(1);
                     printf("\nReplaced node number %d with node number %d \tErrors when grafted node %d: \tErrors when pruned: %d\tErrors when not grafted or pruned: %d \tGRAFTED",NodeNumber,MemNode[k-1],GraftedErrors,PrunedErrors,*UnprunedErrors);
                    *UnprunedErrors=GraftedErrors;
                     Changed=true;
                   }
                 
                  else
                  {
                    T->NodeType=PreviousNodeType;
                    printf("\nNot pruned node number: %d \tNot pruned errors: %d \tPruned Errors: %d ",NodeNumber,*UnprunedErrors,PrunedErrors);

                   }
                 }
             cfree(Temp);
             cfree(MemNode);
            }
else(*MaxNodeNumber)--;
return Changed;
}

     
       
           
           
 
